define(['underscore', 'jquery', 'backbone', 'marionette'], function(_, $, Backbone, Marionette) {
  return Marionette.Controller.extend({
    hasConnectionErrorPopup: false,
    initialize: function(arg) {
      this.app = arg.app;
      this.router = new Backbone.Router;
      this.stack = [];
      this._cache = {};
      this._regionManager = this.getRegionManager();
      this._canvasEl = this.app.canvas.$el[0];
      return this.listenTo(this.app, "device:back", this._triggerCurrentBack);
    },
    fetch: function(id, callback) {
      if ((this.app.patchedViews.hasPatch(id)) && this.app.config.patcher) {
        id = this.app.config.serverURLProtocol + '://' + this.app.config.serverURL + '/mobile/' + this.app.config.version.value + '/views/' + id + '.js';
      } else {
        id = 'views/' + id;
      }
      if (this._cache[id] && !this.app.config.debug) {
        return callback(this._cache[id]);
      } else {
        return require([id], (function(_this) {
          return function(View) {
            _this._cache[id] = View;
            return callback(_this._cache[id]);
          };
        })(this));
      }
    },
    check: function(id) {
      var i, len, ref, view;
      ref = this.stack;
      for (i = 0, len = ref.length; i < len; i++) {
        view = ref[i];
        if (view.id === id) {
          return true;
        }
      }
      return false;
    },
    closeCurrentModals: function() {
      var ref, results;
      results = [];
      while ((((ref = this.current) != null ? ref.screenType : void 0) != null) && this.current.screenType === 'modal') {
        results.push(this.close());
      }
      return results;
    },

    /*
    			Call syntax for .main/dialog/modal/tab:
    				.main(id)
    				.main(id, options)
    				.main(id, callback)
    				.main(id, options, callback)
    
    			id - screens id, screen path without 'views/' prefix and '.js' suffix
    			options - will be passed to view constructor
    			callback - will be invoked after the screen will be loaded
     */

    /*
    			Opens main screen, main screens are available directly from menu
    			to open this type of screen use
    				app.screens.main(...)
    			or link/button with .js-goto-main class and data-screen-id attribute with screen id
    
    			.js-back on that screen opens menu
     */
    main: function(id, options, callback) {
      return this._open('main', id, options, callback);
    },
    show: function(id, options, callback) {
      if (this.app.config.debug) {
        console.error("Error: app.screens.show(...) is deprecated, use app.screens.main(...) instead");
      }
      return this._open('main', id, options, callback);
    },
    getRegionManager: function() {
      return new Marionette.RegionManager();
    },

    /*
    			Opens dialog screen, dialog screens are opened from main screens, they aren't available from menu
    			to open this type of screen use
    				app.screens.dialog(...)
    			or link/button with .js-goto-dialog class and data-screen-id attribute with screen id
    
    			.js-back on that screen returns to previous screen
     */
    dialog: function(id, options, callback) {
      return this._open('dialog', id, options, callback);
    },

    /*
    			Opens modal with overlay, previoys screen is behind the overlay
    			to open this type of screen use
    				app.screens.modal(...)
    			or link/button with .js-goto-modal class and data-screen-id attribute with screen id
    
    				.js-back on that screen closes the modal
     */
    modal: function(id, options, callback) {
      return this._open('modal', id, options, callback);
    },

    /*
    			Used for tabs, screens opened that way inherit type from the screen they are opened from,
    			also they replaces previous screen in screens stack
    			to open this type of screen use
    				app.screens.tab(...)
    			or link/button with .js-goto-tab class and data-screen-id attribute with screen id
    
    				.js-back on that screen behaves the same way, that in screen the current is opened from
     */
    tab: function(id, options, callback) {
      var previous;
      previous = _.last(this.stack);
      previous.view.unlockUI();
      return this._open(previous.options.type, id, options, callback, true);
    },
    replace: function(id, options, callback) {
      return this.tab.apply(this, arguments);
    },
    close: function(callback) {
      var current;
      if (this.stack.length >= 2) {
        if (this.app.config.debug) {
          console.log("Info: Close '" + this.stack[this.stack.length - 1].id + "'");
        }
        current = this.stack.pop();
        current.region.$el.hide();
        current.region.$el.remove();
        current.region.reset();
        if (this._regionManager.get(current.id) != null) {
          this._regionManager.removeRegion(current.id);
        }
        if (current.options.type === 'modal') {
          $(document.body).removeClass('u-ov-h');
        }
        this.current = _.last(this.stack).view;
        this.app.screens.unlockUIAll();
      } else {
        if (this.app.config.debug) {
          console.error("Error: can't go back from screen " + this.current.id + ", there is no screens in screens stack");
          console.log("Info: open menu instead");
        }
        app.menu.open();
      }
      if (callback != null) {
        return callback();
      }
    },
    back: function(callback, params, event, eventArgument) {
      var current, lastView, needReload;
      if (params == null) {
        params = {
          needReload: null
        };
      }
      if (this.stack.length >= 2) {
        if (this.app.config.debug) {
          console.log("Info: back from '" + this.stack[this.stack.length - 1].id + "' to '" + this.stack[this.stack.length - 2].id + "'");
        }
        current = this.stack.pop();
        current.region.$el.hide();
        current.region.$el.remove();
        current.region.reset();
        if (this._regionManager.get(current.id) != null) {
          this._regionManager.removeRegion(current.id);
        }
        if (current.options.type === 'modal') {
          $(document.body).removeClass('u-ov-h o-body-modal');
        }
        lastView = _.last(this.stack);
        lastView.region.$el.show();
        lastView.region.$el.addClass('js-current-screen');
        lastView.view.trigger('screens:back');
        this.current = _.last(this.stack).view;
        if (this.current.lockedUI) {
          this.current.unlockUI();
        }
        if (event) {
          this.current.triggerMethod(event, eventArgument);
        }
        needReload = true;
        if (params.needReload !== undefined && params.needReload !== null) {
          needReload = !!params.needReload;
        } else if (this.current.reloadOnScreensBack !== undefined && this.current.reloadOnScreensBack !== null) {
          needReload = !!this.current.reloadOnScreensBack;
        }
        if (needReload) {
          this.current.checkDataCacheAndReload();
        }
        if (typeof this.router.toFragment === "function") {
          this.router.navigate(this.router.toFragment(_.last(this.stack).id, {
            stack: this._serializeStack()
          }));
        }
        if (callback != null) {
          return callback();
        }
      } else {
        if (this.app.config.debug) {
          console.error("Error: can't go back from screen " + this.current.id + ", there is no screens in screens stack");
          console.log("Info: open menu instead");
        }
        app.menu.open();
        if (callback != null) {
          return callback();
        }
      }
    },
    _open: function(type, id, options, callback, _replace) {
      if (typeof app !== "undefined" && app !== null ? app.config.debug : void 0) {
        app.timer = new Date();
      }
      if (arguments.length === 3 && typeof options === 'function') {
        callback = options;
        options = void 0;
      }
      options = options != null ? options : {};
      return this.fetch(id, (function(_this) {
        return function(View) {
          var previous, region, scrollPosition, view;
          options.app = _this.app;
          options.id = id;
          options.type = type;
          view = new View(options);
          previous = _.last(_this.stack);
          if (previous != null) {
            previous.region.$el.removeClass('js-current-screen');
          }
          region = _this._addRegion(id);
          region.$el.addClass('js-current-screen');
          if (_replace) {
            _this.stack.pop();
            previous.region.$el.hide();
            previous.region.$el.remove();
            previous.region.reset();
            if (_this._regionManager.get(previous.id) != null) {
              _this._regionManager.removeRegion(previous.id);
            }
          } else {
            if (previous && previous.view && previous.view.saveContentScrollPosition) {
              previous.view.saveContentScrollPosition();
            }
            if (type === 'main') {
              _this._removeAll();
              $(document.body).removeClass('u-ov-h o-body-modal');
            } else if (type === 'dialog') {
              _this._hideLast();
              $(document.body).removeClass('u-ov-h o-body-modal');
            } else if (type === 'modal' && (previous != null ? previous.options.type : void 0) === 'modal') {
              previous.region.$el.hide();
            }
          }
          if (type === 'modal') {
            previous.scrollPosition = window.pageYOffset;
            region.$el.addClass('js-modal');
            $(document.body).addClass('u-ov-h o-body-modal');
            view.showOnlyAfterLoadData = true;
          }
          view.screenType = type;
          view.on('ready', _this._onViewReady.bind(_this));
          view.on('render', function(view) {
            return _this.trigger('view:render', view);
          });
          region.show(view);
          region.$el.find('.c-popup').css('max-height', region.$el.innerHeight() - 60);
          scrollPosition = 0;
          _this.stack.push({
            id: id,
            region: region,
            view: view,
            options: options,
            scrollPosition: scrollPosition
          });
          _this.current = view;
          if (typeof _this.router.toFragment === "function") {
            _this.router.navigate(_this.router.toFragment(id, {
              stack: _this._serializeStack()
            }));
          }
          _this.app.menu.close();
          if (_this.app.menu.view) {
            _this.app.menu.view.unlockUI();
          }
          if (_this.app.shortInfo.view) {
            _this.app.shortInfo.view.unlockUI();
          }
          if (callback != null) {
            return callback();
          }
        };
      })(this));
    },
    _triggerCurrentBack: function() {
      var ref;
      return (ref = this.current) != null ? ref.jsBack() : void 0;
    },
    _serializeStack: function() {
      return _.initial(this.stack).map(function(screen) {
        return screen.id;
      });
    },
    _addRegion: function(viewId) {
      var div, region;
      div = document.createElement('div');
      this._canvasEl.appendChild(div);
      region = new Marionette.Region({
        el: div
      });
      this._regionManager.addRegion(viewId, region);
      return region;
    },
    _hideLast: function() {
      var last;
      last = _.last(this.stack);
      last.scrollPosition = window.pageYOffset;
      if (last != null) {
        last.region.$el.hide();
      }
      if (last != null) {
        last.region.$el.removeClass('js-current-screen');
      }
      if ((last != null ? last.region.overlay : void 0) != null) {
        return $(last != null ? last.region.overlay : void 0).hide();
      }
    },
    _removeAll: function() {
      var i, len, previous, ref;
      ref = this.stack;
      for (i = 0, len = ref.length; i < len; i++) {
        previous = ref[i];
        if (previous != null) {
          previous.region.$el.hide();
        }
        if (previous != null) {
          previous.region.$el.remove();
        }
        if ((previous != null ? previous.region.overlay : void 0) != null) {
          $(previous != null ? previous.region.overlay : void 0).remove();
        }
        if (this._regionManager.get(previous.id) != null) {
          this._regionManager.removeRegion(previous.id);
        }
      }
      return this.stack = [];
    },
    _onViewReady: function(view) {
      return this.trigger('view:ready:' + view.id, view);
    },
    unlockUIAll: function() {
      var i, len, ref, ref1, results, view;
      ref = this.stack;
      results = [];
      for (i = 0, len = ref.length; i < len; i++) {
        view = ref[i];
        if (typeof view.unlockUI === "function") {
          view.unlockUI();
        }
        results.push((ref1 = view.view) != null ? ref1.unlockUI() : void 0);
      }
      return results;
    }
  });
});
